@@ -1,3 +1,15 @@
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#      http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
+
 """
     Contains various utilities for `odeint` and numerical methods. Various norms, step size initialization, event callbacks for hybrid systems, vmapped matrix-Jacobian products and some
     additional goodies.
@@ -6,7 +18,7 @@ import attr
 import torch
 import torch.nn as nn
 from torch.distributions import Exponential
-from torchcde import NaturalCubicSpline, natural_cubic_spline_coeffs
+from torchcde import CubicSpline, hermite_cubic_coefficients_with_backward_differences
 
 
 def make_norm(state):
@@ -53,8 +65,8 @@ def adapt_step(dt, error_ratio, safety, min_factor, max_factor, order):
 
 def dense_output(sol, t_sol, t_eval, return_spline=False):
     t_sol = t_sol.to(sol)
-    spline_coeff = natural_cubic_spline_coeffs(t_sol, sol.permute(1, 0, 2))
-    sol_spline = NaturalCubicSpline(t_sol, spline_coeff)
+    spline_coeff = hermite_cubic_coefficients_with_backward_differences(t_sol, sol.permute(1, 0, 2))
+    sol_spline = CubicSpline(t_sol, spline_coeff)
     sol_eval = torch.stack([sol_spline.evaluate(t) for t in t_eval])
     if return_spline:
         return sol_eval, sol_spline
@@ -71,6 +83,7 @@ class EventState:
 
 @attr.s
 class EventCallback(nn.Module):
+    "Basic callback for hybrid differential equations. Must define an event condition and a state-jump"
     def __attrs_post_init__(self):
         super().__init__()
 
@@ -113,3 +126,4 @@ class WrapFunc(nn.Module):
         super().__init__()
         self.f = f
     def forward(self, t, x): return self.f(x)
+
